<?php

namespace InSegment\ApiCore\Models;

use InSegment\ApiCore\Interfaces\SADConsumer;

/**
 * A model of non-recursive iteration of models, their relations and attributes
 */
class SourceAttributeDownstream
{
    const ACCEPT_STATE_INITIAL = 0;
    const ACCEPT_STATE_ACCEPTED = 1;
    const ACCEPT_STATE_SAVED = 2;
    const ACCEPT_STATE_FINAL = 3;
    const ACCEPT_STATE_SOURCE_START = 4;
    const ACCEPT_STATE_DOWNSTREAM_START = 5;
    const ACCEPT_STATE_DOWNSTREAM_PROCESS = 6;
    const ACCEPT_STATE_DOWNSTREAM_END = 7;
    const ACCEPT_STATE_READY_TO_SAVE = 8;
    
    const SOURCE_ITEM_TEMPLATE = [
        'intIndex' => 0,
        'total' => 0,
        'arrayItems' => [],
        'state' => self::ACCEPT_STATE_INITIAL,
        'params' => []
    ];
    
    const ATTRIBUTE_ITEM_TEMPLATE = [
        'intIndex' => 0,
        'total' => 0,
        'arrayItems' => []
    ];
    
    const DOWNSTREAM_ITEM_TEMPLATE = [
        'intIndex' => 0,
        'total' => 0,
        'arrayItems' => []
    ];
    
    /**
     * Stored rules with attributes dictionary
     *
     * @var array 
     */
    protected $classToAttributes = [];
    
    /**
     * Source. It is an array of Models which need to be processes
     * 
     * @var array 
     */
    protected $source = [
        'current' => self::SOURCE_ITEM_TEMPLATE,
        'stack' => []
    ];
    
    /**
     * Attributes. They are retrieved from the dictionary of rules and are iterated on each model
     *
     * @var array 
     */
    protected $attributes = [
        'current' => self::ATTRIBUTE_ITEM_TEMPLATE,
        'stack' => []
    ];
    
    /**
     * Downstream. In the main iteration, it is a separate branch of attributes that will be iterated after the first one
     *
     * @var array 
     */
    protected $downstream = [
        'current' => self::DOWNSTREAM_ITEM_TEMPLATE,
        'stack' => []
    ];
    
    /**
     * Iterating state
     *
     * @var bool 
     */
    protected $isIterating = false;
    
    /**
     * Constructor. Accepts attributes by class dictionary
     * 
     * @param array $classToAttributes
     */
    public function __construct(array $classToAttributes)
    {
        $this->classToAttributes = $classToAttributes;
    }
    
    /**
     * Prepare attribute set
     * 
     * @param array $attributeSet
     * @return array
     */
    public static function prepareAttributeSet(array $attributeSet)
    {
        $result = self::ATTRIBUTE_ITEM_TEMPLATE;
        $result['total'] = count($attributeSet);
        $result['arrayItems'] = $attributeSet;
        return $result;
    }
    
    /**
     * The main cycle. Accepts SADConsumer, array of Models, and a class for attribute rules
     * What the consumer considers 'models', 'attributes' or 'downstream' is irrelevant, also by default
     * the main cycle won't add any new sources or downstream in process, so the consumer should call
     * methods SourceAttributeDownstream::newSource and SourceAttributeDownstream::addDownStream to affect the cycle
     * 
     * Each times the consumer method is called, we stop operating inside that loop and wait for next one.
     * This is done to ensure, that if the consumer method added the new source and the loop is changed, we will
     * react to that immediately.
     * 
     * NOTE: the reason around this "Mighty Mega-function"
     * 
     * ALL the recursion of model relations is done THERE. The array of Models get into Source, they are processed by
     * the Consumer, the Consumer gives new Models and the loop switches to process them on the next iteration. Also
     * they add downstream attributes. Everything else we know beforehand.
     * 
     * The cycle has a switch inside which changes its position in the statements depending on that position
     * and which both affects and is affected by the loop in which it operates. So these entities are tight bound together.
     * 
     * The only two variables the method has in its scope is $this and $consumer, which do not change, but are
     * required on EVERY step of the cycle. So there is no sense to create anything else in visibility.
     * 
     * So we can abstract the manipulation of arrays in properties of $this into methods of $this. But these methods
     * should be protected anyway, because the outside intervention into the arrays in properties of $this is dangerous.
     * And there is no real reason to extend the class, because anything done in another way is not going to work there.
     * So of source we can increase effectiveness of the method calls by declaring the class final,
     * but still the method resolution will be as slow as 7 times incrementing a variable.
     * 
     * Having no extra calls to just manipulate arrays or retrieve values there, we eliminate that overhead completely.
     * And again, the loop in theory works with every model in the collection, with every relation of that model,
     * with every attribute of that model, sometimes twice per attribute, and with every model on the every
     * collection of the every model of the collection, and so on.
     * 
     * If the function grows past 300 lines of code, consider writing another model of non-recursive iteration
     * for the cause which required to write that much.
     * 
     * END NOTE.
     * 
     * @param \InSegment\ApiCore\Interfaces\SADConsumer $consumer
     * @return null
     */
    public function iterate(SADConsumer $consumer)
    {
        if ($this->isIterating) {
            throw new \Exception('SourceAttributeDownstream iteration intervention detected. Iteration should not be called recursively inside another iteration');
        }
        
        $this->isIterating = true;
        
        // as the consumer is allowed to modify state of our cycle, and each of supplied sources must be still processed,
        // method SourceAttributeDownstream::newSource saves the previous state into the stack
        // the outer cycle handles restoration of the state saved into stack once the current source has no more entries
        do {
            // this construction allows to avoid recursion by modification of the state of the cycle
            // the while cycle there iterates through the source, which is possibly changed on almost each iteration
            while ($this->source['current']['intIndex'] < $this->source['current']['total']) {
                // note that on each step consumer's method can and is intended to change OUR cycle's state
                // so we cannot just hop to the next case, after each call of consumer's method we must jump to
                // the while and recheck the state
                switch ($this->source['current']['state']) {
                    
                    // before consumer is informed of model
                    case self::ACCEPT_STATE_INITIAL:
                        if ($this->source['current']['intIndex'] === 0) {
                            $this->source['current']['state'] = self::ACCEPT_STATE_SOURCE_START;
                            $consumer->sourceStarted($this->source['current']['params']);
                        } else {
                            $this->source['current']['state'] = self::ACCEPT_STATE_ACCEPTED;
                            $consumer->acceptModel(
                                $this->source['current']['arrayItems'][$this->source['current']['intIndex']],
                                $this->source['current']['params']
                            );
                        }
                    break;
                    
                    // start of processing source
                    case self::ACCEPT_STATE_SOURCE_START:
                        $this->source['current']['state'] = self::ACCEPT_STATE_ACCEPTED;
                        $consumer->acceptModel(
                            $this->source['current']['arrayItems'][$this->source['current']['intIndex']],
                            $this->source['current']['params']
                        );
                    break;
                
                    // processing of attributes 
                    case self::ACCEPT_STATE_ACCEPTED:
                        if ($this->attributes['current']['intIndex'] < $this->attributes['current']['total']) {
                            // pass attribute to the consumer and advance attribute iteration simultaneously
                            $consumer->acceptAttribute(
                                $this->attributes['current']['arrayItems'][
                                    // micro-optimization: increment index in the same spot where it is used,
                                    // so we don't have to store the result in $attribute variable, and can pass it
                                    // directly to the $consumer::acceptAttribute
                                    $this->attributes['current']['intIndex']++
                                ],
                                $this->source['current']['params']
                            );
                        } else if ($this->downstream['current']['intIndex'] < $this->downstream['current']['total']) {
                            $this->source['current']['state'] = self::ACCEPT_STATE_DOWNSTREAM_START;
                        } else {
                            $this->source['current']['state'] = self::ACCEPT_STATE_READY_TO_SAVE;
                        }
                    break;
                    
                    // inform of start of processing downstream
                    case self::ACCEPT_STATE_DOWNSTREAM_START:
                        $this->source['current']['state'] = self::ACCEPT_STATE_DOWNSTREAM_PROCESS;
                        $consumer->downstreamStarted($this->source['current']['params']);
                    break;
                
                    // process downstream attributes
                    case self::ACCEPT_STATE_DOWNSTREAM_PROCESS:
                        if ($this->downstream['current']['intIndex'] < $this->downstream['current']['total']) {
                            // pass downstream attribute to the consumer and advance downstrem iteration simultaneously
                            $consumer->acceptDownstream(
                                $this->downstream['current']['arrayItems'][
                                    // micro-optimization: increment index in the same spot where it is used,
                                    // so we don't have to store the result in $downstream variable, and can pass it
                                    // directly to the $consumer::acceptDownstream
                                    $this->downstream['current']['intIndex']++
                                ],
                                $this->source['current']['params']
                            );
                        } else {
                            $this->source['current']['state'] = self::ACCEPT_STATE_DOWNSTREAM_END;
                        }
                    break;
                    
                    // inform of end of processing downstream
                    case self::ACCEPT_STATE_DOWNSTREAM_END:
                        $this->source['current']['state'] = self::ACCEPT_STATE_READY_TO_SAVE;
                        $consumer->downstreamEnded($this->source['current']['params']);
                    break;
                
                    // then save of the model
                    case self::ACCEPT_STATE_READY_TO_SAVE:
                        $this->source['current']['state'] = self::ACCEPT_STATE_SAVED;
                        $consumer->saveModel(
                            $this->source['current']['arrayItems'][$this->source['current']['intIndex']],
                            $this->source['current']['params']
                        );
                    break;
                
                    // the model is saved
                    case self::ACCEPT_STATE_SAVED:
                        // advance to the next model and check if the source ends
                        if (++$this->source['current']['intIndex'] === $this->source['current']['total']) {
                            // if this call results in value strictly equal to false, we do not contunue the loop
                            // in that case, break out of the most outer loop (from switch, from while, from do...while)
                            if ($consumer->sourceEnded($this->source['current']['params']) === false) {
                                break 3;
                            }
                        } else {
                            $this->source['current']['state'] = self::ACCEPT_STATE_INITIAL;
                            $this->attributes['current']['intIndex'] = 0; // reset attributes iteration
                        }
                    break;
                }
            }
        } while (
            // if some source is stacked, we restore the stacked source, otherwise the loop will exit
            $this->unstackSource()
        );
        
        $this->isIterating = false;
    }
    
    /**
     * Add new array of Models so it will become the current source and stack everything down
     * 
     * @param array $source
     * @param string $class
     * @param array $params
     * @return null
     */
    public function newSource(array $source, string $class, array $params = [])
    {
        $this->attributes['stack'][] = $this->attributes['current'];
        $this->attributes['current'] = $this->classToAttributes[$class];
        
        $this->downstream['stack'][] = $this->downstream['current'];
        $this->downstream['current'] = self::DOWNSTREAM_ITEM_TEMPLATE;
        
        $this->source['stack'][] = $this->source['current'];
        $this->source['current'] = self::SOURCE_ITEM_TEMPLATE;
        $this->source['current']['params'] = $params;
        $this->source['current']['total'] = count($source);
        $this->source['current']['arrayItems'] = &$source;
    }
    
    /**
     * Add something to the end of the downstream
     * 
     * @param mixed $value
     */
    public function addDownStream($value)
    {
        $this->downstream['current']['arrayItems'][] = $value;
        ++$this->downstream['current']['total'];
    }
    
    /**
     * Ladder up the stack if possible, return true if successful, false if the stack was empty
     * 
     * @return boolean
     */
    protected function unstackSource()
    {
        if (empty($this->source['stack'])) {
            return false;
        }
        
        $this->source['current'] = array_pop($this->source['stack']);
        $this->downstream['current'] = array_pop($this->downstream['stack']);
        $this->attributes['current'] = array_pop($this->attributes['stack']);
        return true;
    }
}
