<?php

namespace InSegment\ApiCore\Services;

use Illuminate\Support\Facades\App;
use Illuminate\Support\Facades\Cache;
use InSegment\ApiCore\Services\MicroEvent;

class AttributeCache
{
    const CACHE_HAS_CAST = 1 << 1;
    const CACHE_HAS_METHOD = 1 << 2;
    const CACHE_BASE_HAS_METHOD = 1 << 3;
    const CACHE_HAS_GET_MUTATOR = 1 << 4;
    const CACHE_HAS_SET_MUTATOR = 1 << 5;
    const CACHE_IS_DATE_ATTRIBUTE = 1 << 6;
    const CACHE_IS_JSON_CASTABLE = 1 << 7;
    const CACHE_FILL_JSON = 1 << 8;
    const CACHE_IN_DATES = 1 << 9;
    
    /**
     * Prefix for cache
     *
     * @var string 
     */
    protected static $prefix;
    protected static $cache;
    protected static $castsCache = [];
    protected static $castTypeCache = [];
    protected static $dateFormatCache = [];
    protected static $isHandlerSet = false;
    protected static $isDownForMaintenance = false;
    
    /**
     * Try to use cache for attribute of models of $class
     * 
     * @param string $class
     * @param string $boxName
     * @param string $attribute
     * @return int|null Bitmask
     */
    public static function try(string $class, string $boxName, string $attribute)
    {
        return static::$cache[$class][$boxName][$attribute] ?? (!isset(static::$cache) ? static::tryBoot($class, $boxName, $attribute) : null);
    }
    
    /**
     * Try to use cache for attribute casts of models of $class
     * 
     * @param string $class
     * @param string $boxName
     * @param string $attribute
     * @return array|null
     */
    public static function tryCastType(string $class, string $boxName, string $attribute)
    {
        return static::$castTypeCache[$class][$boxName][$attribute] ?? null;
    }
    
    /**
     * Store cast type for attribute of models of $class
     * 
     * @param string $castType
     * @param string $class
     * @param string $boxName
     * @param string $attribute
     * @return null
     */
    public static function cacheCastType($castType, string $class, string $boxName, string $attribute)
    {
        static::$castTypeCache[$class][$boxName][$attribute] = $castType;
    }
    
    /**
     * Try to use cache for casts of models of $class
     * 
     * @param string $class
     * @param string $boxName
     * @return array|null
     */
    public static function tryCasts(string $class, string $boxName)
    {
        return static::$castsCache[$class][$boxName] ?? null;
    }
    
    /**
     * Store casts for models of $class
     * 
     * @param array $casts
     * @param string $class
     * @param string $boxName
     * @return null
     */
    public static function cacheCasts($casts, string $class, string $boxName)
    {
        static::$castsCache[$class][$boxName] = $casts;
    }
    
    /**
     * Try to get from the store the date format for conneciton name
     * 
     * @param string|null $connectionName
     * @param mixed $format
     */
    public static function tryDateFormat(string $connectionName = null)
    {
        return static::$dateFormatCache[$connectionName] ?? null;
    }
    
    /**
     * Store date format for connection name
     * 
     * @param mixed $format
     * @param string $connectionName
     */
    public static function cacheDateFormat($format, string $connectionName = null)
    {
        static::$dateFormatCache[$connectionName] = $format;
    }
    
    /**
     * Cache characteristics for an attribute of models of $class
     * 
     * @param string $class
     * @param string $box
     * @param string $attribute
     * @param int $charateristicsMask
     * @throws \Exception
     */
    public static function cache(string $class, string $box, string $attribute, int $charateristicsMask)
    {
        if (static::$isDownForMaintenance) {
            return;
        }
        
        static::$cache[$class][$box][$attribute] = $charateristicsMask;
        
        if (!static::$isHandlerSet) {
            MicroEvent::registerEvent('InSegment\ApiCore\Middleware\TerminationEvent', 'termination', function () {
                if (config('api_core.attibutes_cache')) {
                    Cache::forever(static::getCachePrefix(), static::$cache);
                }
            });
            
            static::$isHandlerSet = true;
        }
    }
    
    /**
     * Try to boot cache from the caching storage and use cache for attribute of models of $class
     * 
     * @param string $class
     * @param string $box
     * @param string $attribute
     * @return int|null Bitmask
     */
    protected static function tryBoot(string $class, string $box, string $attribute)
    {
        if (static::$isDownForMaintenance) {
            return null;
        } else if (config('api_core.attibutes_cache')) {
            static::$cache = Cache::get(static::getCachePrefix()) ?? [];
            if ((static::$isDownForMaintenance = App::isDownForMaintenance())) {
                return null;
            }
        } else {
            static::$cache = [];
        }
        
        return static::$cache[$class][$box][$attribute] ?? null;
    }
    
    /**
     * Get cache prefix
     * 
     * @return string
     */
    protected static function getCachePrefix()
    {
        if (!isset(static::$prefix)) {
            static::$prefix = config('api_core.cache_prefix', 'insegment.api-core.') . 'attribute-cache';
        }
        
        return static::$prefix;
    }

}
